package sqlite

import (
	"database/sql"
	"time"

	ippb "git.autistici.org/ai3/tools/iprep/proto"
)

var statements = map[string]string{
	"insert_event": `
INSERT INTO events (ip, event_type, count, timestamp) VALUES (?, ?, ?, ?)
`,
	"scan_by_ip": `
SELECT
  event_type, SUM(count) AS sum
FROM
  events
WHERE
  ip = ?
  AND timestamp > ?
GROUP BY
  event_type
`,
}

type DB struct {
	db    *sql.DB
	stmts StatementMap
}

func Open(path string) (*DB, error) {
	db, err := sqlOpen(path)
	if err != nil {
		return nil, err
	}

	stmts, err := NewStatementMap(db, statements)
	if err != nil {
		db.Close()
		return nil, err
	}

	return &DB{
		db:    db,
		stmts: stmts,
	}, nil
}

func (db *DB) Close() {
	db.db.Close()
}

func (db *DB) AddAggregate(aggr *ippb.Aggregate) error {
	tx, err := db.db.Begin()
	if err != nil {
		return err
	}
	stmt := db.stmts.Get(tx, "insert_event")
	now := time.Now()

	for _, bt := range aggr.ByType {
		for _, item := range bt.ByIp {
			if _, err := stmt.Exec(item.Ip, bt.Type, item.Count, now); err != nil {
				tx.Rollback() // nolint
				return err
			}
		}
	}

	return tx.Commit()
}

func (db *DB) WipeOldData(age time.Duration) error {
	_, err := db.db.Exec(`
DELETE FROM events WHERE timestamp < ?
`, time.Now().Add(-age))
	return err
}

func (db *DB) ScanIP(startTime time.Time, ip string) (ippb.Map, error) {
	tx, err := db.db.Begin()
	if err != nil {
		return nil, err
	}
	defer tx.Rollback() // nolint

	rows, err := db.stmts.Get(tx, "scan_by_ip").Query(ip, startTime)
	if err != nil {
		return nil, err
	}
	defer rows.Close()

	m := make(ippb.Map)
	for rows.Next() {
		var evType string
		var count int64
		if err := rows.Scan(&evType, &count); err != nil {
			return nil, err
		}
		m.Incr(evType, ip, count)
	}

	return m, nil
}

func (db *DB) ScanType(startTime time.Time, t string) (ippb.Map, error) {
	return nil, nil
}
